/**/

   /*  Notes:                                                                                 */ /*fold00*/
   /* CRON.CMD -- cheap UNIX cron clone for OS/2
      Jeroen Hoppenbrouwers (hoppie@kub.nl), October 1994
      Bug fixes by Paul Ratcliffe, October 2003
      More bug fixes and additions by Anton Monroe <amonroe5301@gmail.com>
      May 2006, November 2008, January 2011, July 2018, July 2019, July 2023
   */

   /***************************************************************************************************/
   /* Notes by Anton Monroe, May 2006, November 2008, January 2011, July 2018, July 2019              */
   /*
      USAGE:
             CRON.CMD [options]
      or more often, from STARTUP.CMD
             detach CRON.CMD [options]
      or even better, from the StartCron.cmd script
             StartCron.cmd START
             StartCron.cmd STOP
             etc.
      for options and defaults, type "StartCron.cmd help"

      I assume you are somewhat familiar with the Unix type of cron daemon (Vixie cron)
      and crontab file. This Cron.cmd and the crontab file include many features of other Crons,
      but not all of them:
        * There is no support for multiple users.  Unix cron uses a confusing
          variety of crontab files-- user crontab files in /var/spool/cron/tabs/,
          a system-wide crontab file in /etc/, and other system crontabs in /etc/cron.d/.
          The system crontabs have an extra field to specify which user a command
          will be run as.
          This Cron.cmd uses only one crontab file, which does not have a user field.

        * Unlike Unix cron, you cannot set environment variables like SHELL or
          PATH in the crontab. Instead, you can use an entry like "DoSomething.cmd"
          and have DoSomething.cmd set whatever environment it needs.

        * Unix cron can mail a user if there is a problem.  Cron.cmd cannot.
          It would be a good thing to add. As with the environment, you could write a
          DoSomething.cmd that provides its own mailer.

          You could pipe stderr to a separate mailer, with a crontab entry like this:
              (4OS2.exe /c DoSomething.cmd) >nul |& cronmail.cmd -subject="Error from DoSomething.cmd"

        * This version of Cron.cmd now supports ranges, lists, step values, and
          names for months and days of the week.


      Each line of the crontab file has six fields:
          minute       hour      day-of-month      month       day-of-week       command-to-run
      Minutes range from 0-59, hours from 0-23, days from 1-31, months from 1-12,
      day of week from 1-7, where Monday is 1 and Sunday is 7.
      The month and day-of-week can also be specified by names (in English).
      Only the first three letters of the name are significant.

      Values are specified as in other Crons. So in the 'month' field
          1,2,3,4       matches January, February, March, or April
          1-4           means the same
          Jan-Apr       means the same
          1-6/2         matches even-numbered months between Jan and June. That is, Feb, Apr, and Jun
          * /2          means every second month; Feb, Apr, Jun, Aug, Oct, Dec
          *             matches any month

      So, to run FOO.CMD every two hours between 08:00 and 18:00, only on Monday-Friday:
            0 8,10,12,14,16,18 * * 1-5 foo.cmd
      or a shorter way to write it:
            0 8-18/2 * * 1-5 foo.cmd
            0 8-18/2 * * Mon-Fri foo.cmd

      ==> Note that the start of a range must be less than the end.  "Sun-Tue" is the
      same as "7-2", which is not a valid range. You must use "Sun,Mon,Tue".
   */

   /*
      Fixes by akm:
         * previous version stopped reading crontab at the first blank line;
           it now ignores blank lines and keeps reading.
         * fixed day-of-month / day-of-week logic to be compatible with other crons.
           So now a line like
               * * 3 * 5 foo.exe
           will run foo.exe if today is EITHER the third day of the month OR
           the fifth day of the week.

      Added a lot of things, but it is still as cheap as ever:
         * Many comments, because that's the way I write.

         * Added support for step values, Month and Day-of-week names.

         * Changed default crontab to %etc%\crontab or %unixroot%\etc\crontab

         * Added ability to pass options on the command line. But note that the
           'getopts' routine is not a clone of the Unix getopts.

         * Added some simple handlers for Syntax, NoValue, etc., so
           when the detached session crashes it will leave some hint in the
           logfile about what happened.

         * added some error messages and debugging messages.  Verbosity
           level can be controlled by the -debug switch. Most messages have a
           [<something>] prefix, which someday could be used to give better control
           over which messages are printed.

         * This Cron.cmd writes directly to a log file. Previously, output
           was to stdout.  Logging had to be done by redirecting stdout to a
           file, which kept the log file always open.
           If you prefer the old behavior, change the value of the
           'Opt.0stdout' variable or use "Cron.cmd -stdout". (It should work.
           Since I do not use stdout, I have not tested it in a long time.)

         * Cron.cmd can change the name of the logfile on the first of each
           month, or each day, so I do not need a separate utility to rotate
           log files.

         * Added a script that can control the Cron.cmd daemon while it is running.
           Rather like the /etc/init.d/Start* scripts in Linux. So you can type
               StartCron.cmd START     to start Cron.cmd
               StartCron.cmd STOP      to stop it
               StartCron.cmd STATUS    to find out if it is running
               StartCron.cmd PAUSE     to temporarily suspend running jobs
           type "StartCron.cmd -help" for more.
           This also required writing a PID file and creating a Rexx queue to
           communicate with the daemon.

         * Improved the organization by moving code into subroutines. Limited
           the number of shared variables by using 'procedure' when possible.

         * Added a 'pause/resume' feature.  When Paused, Cron.cmd will not
           launch any jobs.  Instead it will keep a list of jobs that were
           skipped, then start running them when the pause time expires, or when
           it receives a 'Resume' signal.
   */

   /*
      akm: things to fix:
           * either remove the -stdout option, or provide a way to do it from StartCron.cmd
           * add support for @reboot, @hourly, @daily, etc.
           * when paused, let "startcron.cmd status" show what the delayed jobs are,
             not just how many.  Maybe by saving them in a text file.
           * I should write a separate syntax-checking routine to verify crontab entries when
             crontab is read.
           * the error-handling is primitive. What kind of errors is a Cron daemon
             likely to have?
   */
   /***************************************************************************************************/

   /*
      Some configuration variabls:
         Opt.0debug        = 0       --normal logging level
         Opt.0debug        = 1       --adds a few messages used for debugging.
         Opt.0debug        = 2       --adds a few more messages
         Opt.0debug        = 3       --adds a few more messages
         Opt.0renamelog    = 'NEVER' --use the same logfile forever.
         Opt.0renamelog    = 'MONTH' --change the logfile name every month.
         Opt.0renamelog    = 'DAY'   --change the logfile name every day.
         Opt.0stdout       = 0       --write messages to the log file.
         Opt.0stdout       = 1       --write messages to stdout instead of a log file.

     Note that Opt.0stdout = 0 means the logfile will not include the messages about
     PIDs when something is detached, or some error messages. If you do not like
     that, use '-stdout' and redirect Cron.cmd's output to a log file.
   */

   Init: /*fold00*/
   /* =============================================================================================== */
   /*  initialize a lot of variables, read command line                                               */
   /* =============================================================================================== */
   /* Load RexxUtil and RXU libraries                                                                 */
   if RxFuncQuery('SysLoadFuncs') then do
         call RxFuncAdd 'SysLoadFuncs', 'RexxUtil', 'SysLoadFuncs'
         call SysLoadFuncs
      end
   if RxFuncQuery('rxqprocstatus') then
      call RxFuncAdd 'rxuinit', 'rxu', 'rxuinit'
   call RxuInit

   signal on Syntax
   signal on NoValue
   call on Error
   call on Failure
   call on Notready

   PauseFor = 0                                                          /* time left before resuming */
   Opt. = ''                                                            /* holds command line options */
   Delayed. = ''                                            /* holds list of delayed jobs when Paused */
   Delayed.0 = 0
   Cron. = ''
   /*    default values                                                                               */
   Opt.0debug = 0
   Opt.0renamelog = 'MONTH'
   Opt.0stdout = 0

   /*    Shared is a list of important variables that are likely to be needed by many procedures      */
   Shared = 'G. Opt. entry. maxN Delayed. Cron. PauseFor'

   G. = ''
   parse upper source . . G.0meLong
   G.0me = filespec('name', G.0meLong)

   parse arg args
   interpret GetOpts(args)

   /*    ResolveOptions sets variables based on command line and defaults                             */
   call ResolveOptions
   if Opt.0help \== '' then do
         call Help
         return 1
      end

   /*  first make sure we can write to the log file. Everything else depends on it                    */
   /*  The only time we use stderr or stdout is if the log file is not available                      */
   call off NotReady
   ret = lineout(Cron.0logfile)
   if ret \== 0 then do
         call lineout 'stderr', 'Error at line' LineNo()
         call lineout 'stderr', 'Cannot write to logfile' Cron.0logfile
         call lineout 'stderr', 'Aborting'
         return 2
      end

   ret = lineout(Cron.0pidfile)
   if ret \== 0 then do
         call LogIt '[ERROR] Cannot write to pidfile' Cron.0pidfile
         call LogIt "[ERROR] there will be no PID file"
         Cron.0pidfile = ''
      end
   call on Notready

   /*    create PID file and control queue.  They should not exist yet, but make sure                 */
   if Cron.0pidfile \= '' then call SysFileDelete Cron.0pidfile
   call UpdatePIDfile
   if Cron.0qname \= '' then do
         call rxqueue 'delete', Cron.0qname
         call rxqueue 'create', Cron.0qname
         oldQname = rxqueue('set', Cron.0qname)
      end

   /* ----------------------------------------------------------------------------------------------- */
   /*   make sure we have everything we need                                                          */
   if Cron.0crontab == '' then do
         call LogIt '[ERROR] Cannot find a crontab file. Quitting.'
         call Egress
      end
   if \FileExist(Cron.0crontab) then do
         call LogIt '[ERROR] Cannot find the crontab file:' Cron.0crontab'. Quitting.'
         call Egress
      end
   /* ----------------------------------------------------------------------------------------------- */

   crondate.old = stream(Cron.0crontab, 'c', 'query datetime')
   call LogIt '==============================================================================='
   call LogIt '[START]' G.0me 'using 'Cron.0crontab
   call ReadCrontab
   /* === End of Init =============================================================================== */

   Main_Loop: /*fold00*/
   /* =============================================================================================== */
   /*    (1) Check control queue for instructions                                                     */
   /*    (2) Check the crontab date; if new, re-read it.                                              */
   /*    (3) for each command, if the time matches then run command                                   */
   /*    (4) sleep until end of the minute, then start over                                           */
   /* =============================================================================================== */
   do forever
      call CheckQueue
      crondate.new = stream(Cron.0crontab, 'c', 'query datetime')
      if Opt.0debug > 2 then
         call LogIt '[DEBUG] crondate.old='crondate.old'  crondate.new='crondate.new
      if crondate.new \= crondate.old then do
            call LogIt '[RELOAD] new' Cron.0crontab 'dated' crondate.new
            call ReadCrontab
            crondate.old = crondate.new
         end
      call RunThings
      call SysSleep 60 - (time('s') - time('m') * 60)
   end
   call Egress
   /* === End of Main_Loop ========================================================================== */

   CheckQueue: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    check queue for control commands and executes them                                           */
   /* =============================================================================================== */
   do while queued() > 0
      ControlCmd = strip(linein('QUEUE:'))
      call LogIt '[QUEUE] received from queue: 'ControlCmd
      parse var ControlCmd Action ActionArgs
      Action = translate(Action)
      select
         when Action == 'STOP' then do
               call Egress
            end
         when abbrev(Action, 'PAUSE') == 1 then do
               /*  if no pause time is given, default to 60 minutes                                   */
               if ActionArgs == '' then ActionArgs = 60

               if \datatype(ActionArgs, 'w') then
                  call LogIt '[ERROR] line 'LineNo()' invalid Pause value "'ActionArgs'"'
               else do
                     /*  add 1 because RunThings will decrement Pause right after this                */
                     PauseFor = ActionArgs +1
                     call LogIt '[PAUSE] cron will pause jobs for' PauseFor -1 'minutes'
                     call UpdatePIDfile
                  end
            end
         when Action == 'RESUME' then do
               PauseFor = 0
               call LogIt '[PAUSE] cron has resumed running jobs'
               call UpdatePIDfile
            end
         when Action == 'DEBUG' then do
               if \datatype(ActionArgs, 'w') then
                  call LogIt '[ERROR] line 'LineNo()' invalid Debug value "'ActionArgs'"'
               else do
                     Opt.0debug = ActionArgs
                     call UpdatePIDfile
                  end
            end
         otherwise
            call LogIt '[ERROR] line 'LineNo()' unknown queue command'
      end   /*  end SELECT  */
   end   /*  end DO  */
   return
   /* === End of CheckQueue ========================================================================= */

   LogCommandList: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    prints the command list to the log (if Opt.0debug is set)                                    */
   /* =============================================================================================== */
   call lineout Cron.0logfile, 'Found these commands in 'Cron.0crontab
   fldlen = 7
   logstr = ' 'right('min:', fldlen)||right('hr:', fldlen)||right('day:', fldlen) ,
            ||right('mnth:', fldlen)||right('dow:', fldlen)||'    command:'
   call lineout Cron.0logfile, logstr
   do N = 1 to maxN
      logstr = right(entry.N.minute, fldlen)||right(entry.N.hour, fldlen)||right(entry.N.DoM, fldlen) ,
               ||right(entry.N.month, fldlen)||right(entry.N.DoW, fldlen)||'     'entry.N.command
      call lineout Cron.0logfile, logstr
   end
   call lineout Cron.0logfile, 'End of list'
   call lineout Cron.0logfile
   return
   /* === End of LogCommandList ===================================================================== */

   RunThings: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /* checks each set of values; if they match the current time, run the command                      */
   /* =============================================================================================== */
   /*    fix: separate this into two functions-- (1) decide what to run, then                         */
   /*                                            (2) call a sub to run, errorcheck, and log           */
   NeedPID = 0
   currMin = time('m') - time('h') * 60
   currHour = time('h')
   today = date('o')
   parse var today currYear '/' currMonth '/' currDay
   today = date('w')
   currDoW = wordpos(today, 'Monday Tuesday Wednesday Thursday Friday Saturday Sunday')

   /*  decrement Pause timer by 1 minute                                                              */
   if PauseFor > 0 then do
         PauseFor = PauseFor -1
         NeedPID = 1
      end

   /*  If there is a backlog of delayed jobs, we do not want to start a flurry of activity            */
   /*  by running them all at once. We cannot know how long each job will take or how much            */
   /*  of the cpu it needs, but waiting one minute between delayed jobs should at least help.         */
   if Delayed.0 > 0 ,
    & PauseFor == 0 then do
         call LogIt '[RUN] run Delayed job "'Delayed.1'"'
         '@detach' Delayed.1
         ret = SysStemDelete('Delayed.', 1, 1)
         if ret \= 0 then
            call LogIt '[ERROR] line 'LineNo()' SyStemDelete returned 'ret
         NeedPID = 1
      end

   do N = 1 to maxN
      if Opt.0debug > 2 then
         call LogIt ,
              '[DEBUG] RunThings checking "'entry.N.minute   entry.N.hour   entry.N.DoM   entry.N.month   entry.N.DoW'"'
      isMin   = MatchField(entry.N.minute, currMin, N)
      isHour  = MatchField(entry.N.hour, currHour, N)
      isMonth = MatchField(entry.N.month, currMonth, N)
      /*     if any of the above == 0 then forget it                                                  */
      if isMin + isHour + isMonth \= 3 then iterate
      isDoM   = MatchField(entry.N.DoM, currDay, N)
      isDoW   = MatchField(entry.N.DoW, currDoW, N)
      /*  Normally, a cron job runs only if ALL of the entry fields match the current time, so        */
      /*          0 0 1 * * <command>                                                                 */
      /*  will run if Minute == 0, AND Hour == 0, AND DoM == 1, in any month on any Day-of-Week.      */
      /*  An inconsistent but useful feature of Unix cron is that if a Day-of-Month and Day-of-Week   */
      /*  are both specified, the job will run if EITHER the Day-of-Month OR the Day-of-Week matches. */
      /*            0 0 1 * 1 <command>                                                               */
      /*    will run if today is the first of a month OR if today is a Monday                         */
      if entry.N.DoM \= '*' ,
       & entry.N.DoW \= '*' then
         RunIt = (isMin ,
                 & isHour ,
                 & isMonth ,
                 & (isDoM ,
                 | isDoW))
      else
         RunIt = (isMin ,
                 & isHour ,
                 & isMonth ,
                 & isDoM ,
                 & isDoW)
      if RunIt then do
            if entry.N.command = 'renamelog' then do
                  if Opt.0debug > 1 then
                     call LogIt '[DEBUG] 'entry.N.minute   entry.N.hour   entry.N.DoM   entry.N.month   entry.N.DoW
                  if Opt.0debug > 1 then
                     call LogIt '[DEBUG] 'currMin   currHour   currDay   currMonth   currDoW
                  call LogIt '[LOG] close old log' Cron.0logfile
                  call SetLogfileName
                  call LogIt '[LOG] start new log' Cron.0logfile
                  NeedPID = 1
                  iterate
               end
            if PauseFor == 0 then do
                  call LogIt '[RUN] 'entry.N.command
                  if Opt.0debug > 1 then call LogIt '[DEBUG] N='N
                  if Opt.0debug > 1 then
                     call LogIt '[DEBUG] 'entry.N.minute   entry.N.hour   entry.N.DoM   entry.N.month   entry.N.DoW
                  if Opt.0debug > 1 then
                     call LogIt '[DEBUG] 'currMin   currHour   currDay   currMonth   currDoW
                  '@detach' entry.N.command
               end
            else do
                  /* add entry.N.command to list of delayed jobs, unless it duplicates one already listed */
                  if Opt.0debug > 0 then
                     call LogIt '[RUN] skip job "'entry.N.command'" because Paused'
                  Dupe = 0
                  do Dctr = 0 to Delayed.0
                     if Delayed.Dctr == entry.N.command then do
                           Dupe = 1
                           if Opt.0debug > 0 then
                              call LogIt '[RUN] do not save duplicate delayed job "'entry.N.command'"'
                           leave
                        end
                  end   /*  end DO Dctr =  */
                  if Dupe == 0 then do
                        DHigh = Delayed.0 +1
                        Delayed.DHigh = entry.N.command
                        Delayed.0 = DHigh
                        NeedPID = 1
                        if Opt.0debug > 0 then
                           call LogIt '[RUN] save job "'entry.N.command'" for later'
                        if Opt.0debug > 0 then
                           call LogIt '[RUN] There are 'Delayed.0 ' commands saved for later'
                     end   /*  end if Dupe == 0  */
               end
         end   /*  end if RunIt  */
   end   /*  end DO N = 1 to maxN  */
   if NeedPID then do
         call UpdatePIDfile
         NeedPID = 0
      end
   return
   /* === End of RunThings ========================================================================== */

   MatchField: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    returns 1 if a time field (arg 1) is a match for the current value (arg 2)                   */
   /*  This is called by RunThings.                                                                   */
   /* =============================================================================================== */
   fld = arg(1)
   TimeNow = arg(2)
   N = arg(3)
   if Opt.0debug > 2 then call LogIt '[DEBUG] MatchField checking "'fld'" at 'TimeNow
   /*  fld may be several values separated by commas. For each value, try a series of tests.          */
   /*  each test either returns 1 or iterates to the top of loop.                                     */
   do while fld \= ''
      /*     get one comma-separated value and split off any step value.                              */
      /*     Note that a range can also have a step value                                             */
      parse var fld Thisvalue ',' fld
      parse var Thisvalue Thisvalue '/' step

      if Thisvalue = '*' then do
            select
               when step == '' then
                  return 1
               when (TimeNow // step) = 0 then
                  return 1
               otherwise
                  iterate
            end
         end

      parse var Thisvalue Timefrom '-' Timeto
      if Opt.0debug > 2 then call LogIt '[DEBUG] MatchField line' LineNo() 'Thisvalue='Thisvalue 'Timefrom='Timefrom 'Timeto='Timeto 'TimeNow='TimeNow
      /*    it is single number, not a range                                                          */
      if Timeto == '' then do
            if Thisvalue = TimeNow then
               return 1
            else
               iterate
         end

      /*    it is a range. Check for reversed values                                                  */
      if Timefrom > Timeto then do
            call LogIt '[ERROR] line 'LineNo()', invalid range: "'Timefrom'-'Timeto'" in 'Cron.0crontab
            call LogIt '[ERROR] "'CrontabEntry(N)'"'
            iterate
         end

      if Timefrom <= TimeNow ,
       & TimeNow <= Timeto then do
            if Opt.0debug > 2 then call LogIt '[DEBUG] MatchField line' LineNo() 'Timefrom='Timefrom 'TimeNow='TimeNow 'Timeto='Timeto
            if step == '' then return 1
            if (TimeNow // step) = 0 then return 1
         end
   end   /*  end DO while fld \= ''  */
   return 0
   /* === End of MatchField ========================================================================= */

   CrontabEntry: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    returns the complete crontab entry number N                                                  */
   /*    because it might be useful                                                                   */
   /*    fix: it would be better to return the original as in crontab, not converted to numbers       */
   /* =============================================================================================== */
   parse arg N
   return entry.N.minute entry.N.hour entry.N.DoM entry.N.month entry.N.DoW  entry.N.command
   /* === End of CrontabEntry ======================================================================= */

   ReadCrontab: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  reads crontab commands into an array.  The crontab is only re-read if it changes.              */
   /*  each crontab command number N is composed of                                                   */
   /*     entry.N.minute                                                                              */
   /*     entry.N.hour                                                                                */
   /*     entry.N.DoM                                                                                 */
   /*     entry.N.month                                                                               */
   /*     entry.N.dow                                                                                 */
   /*     entry.N.command                                                                             */
   /* ----------------------------------------------------------------------------------------------- */
   /*  I want to start a new log at the beginning of each month.  The easiest way to do that is       */
   /*  to treat it like any other job that cron does, by adding it to the list of commands.           */
   /* =============================================================================================== */
   N = 0                                                                   /* number of crontab entry */
   Thisline = linein(Cron.0crontab, 1, 0)
   do while chars(Cron.0crontab) > 0
      Thisline = linein(Cron.0crontab)
      if left(Thisline, 1) = '#' then iterate
      if Thisline == '' then iterate
      else do
            N = N +1
            entry.N. = 0
            parse var Thisline entry.N.minute entry.N.hour entry.N.DoM entry.N.month entry.N.DoW entry.N.command
            entry.N.month = MonthToNum(entry.N.month, N)
            entry.N.DoW = DayToNum(entry.N.DoW, N)
         end
   end
   call stream Cron.0crontab, 'c', 'close'

   /* if log file is to be renamed, add a fake crontab entry for it                                   */
   select
      when Opt.0renamelog == 'DAY' then do
            Thisline = '0 0 * * * renamelog'
            N = N +1
            entry.N. = 0
            parse var Thisline entry.N.minute entry.N.hour entry.N.DoM entry.N.month entry.N.DoW entry.N.command
            maxN = N
         end
      when Opt.0renamelog == 'MONTH' then do
            Thisline = '0 0 1 * * renamelog'
            N = N +1
            entry.N. = 0
            parse var Thisline entry.N.minute entry.N.hour entry.N.DoM entry.N.month entry.N.DoW entry.N.command
            maxN = N
         end
      when Opt.0renamelog == 'NEVER' then
         nop
      otherwise
         call LogIt '[ERROR] invalid renamelog value "'Opt.0renamelog'"'
   end
   if Opt.0debug > 0 then call LogCommandList
   return
   /* === End of ReadCrontab ======================================================================== */

   UpdatePIDfile: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  akm: As is usual for a PID file, the first line of the PID file is the PID of                  */
   /*  the Cron daemon, but I also add a second line that shows some parameters, so                   */
   /*  StartCron.cmd can easily check the status of the Cron daemon                                   */
   /*                                                                                                 */
   /*  Note that we use lineout() to set the write position to line 1 and overwrite                   */
   /*  the old contents of the file.  So the file's 'Modify' time is updated but the                  */
   /*  'Create' time does not change.  "StartCron.cmd status" uses the Create time to                 */
   /*  report when the daemon was started.                                                            */
   /*                                                                                                 */
   /*  There is a harmless but slightly confusing side effect-- The old contents of                   */
   /*  the file are not erased, just overwritten with two lines, each terminated with                 */
   /*  a CRLF.  If the new text is shorter than the old text, some of the old text                    */
   /*  will remain as a third line. StartCron.cmd will only read the first two lines.                 */
   /* =============================================================================================== */
   if Cron.0pidfile \= '' then do
         if Opt.0debug > 0 then call LogIt '[DEBUG] updating PID file 'Cron.0pidfile
         procinfo = RxProcId()
         parse var procinfo pid .
         Cronargs = 'logfile='Cron.0logfile
         Cronargs = Cronargs' crontab='Cron.0crontab
         Cronargs = Cronargs' pidfile='Cron.0pidfile
         Cronargs = Cronargs' Queue='Cron.0qname
         Cronargs = Cronargs' RenameLog='Opt.0renamelog
         Cronargs = Cronargs' debug='Opt.0debug
         Cronargs = Cronargs' Pause for='PauseFor
         Cronargs = Cronargs' Delayed jobs='Delayed.0
         call lineout Cron.0pidfile, pid, 1
         ret = lineout(Cron.0pidfile, Cronargs)
         call lineout Cron.0pidfile
         if ret \= 0 then call LogIt '[ERROR] [UpdatePIDfile] line 'LineNo()' ret == 'ret
      end
   return
   /* === End of UpdatePIDfile ====================================================================== */

   ResolveOptions: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    determines names of files and other options, based on commandline and defaults               */
   /*    Cron.0logdir      == commandline  |  %unixroot\var\log\     |  current directory             */
   /*    Cron.0crontab     == commandline  |  %etc\crontab           |  %unixroot\etc\crontab         */
   /*    Cron.0pidfile     == commandline  |  %unixroot\var\run\cron.pid                              */
   /*    Cron.0qname       == commandline  |  CRON_CONTROL                                            */
   /*    Opt.0renamelog    == commandline  |  'MONTHLY'                                               */
   /*    Opt.0debug        == commandline  |  '0'                                                     */
   /* =============================================================================================== */

   unixroot = value('unixroot', , 'os2environment')
   etc = value('etc', , 'os2environment')
   /*    first get the names of logfiles and the debug value, so we can log error messages            */
   if Opt.0logdir \= '' then
      Cron.0logdir = Opt.0logdir
   else do
         if unixroot \== '' then
            Cron.0logdir = unixroot'\var\log\'
         else
            Cron.0logdir = '.\'
      end
   /*  make sure logdir has a trailing \                                                              */
   Cron.0logdir = strip(translate(Cron.0logdir, '\', '/'), 't', '\')||'\'

   if Opt.0stdout then
      Cron.0errorlogfile = 'stdout'
   else
      Cron.0errorlogfile = Cron.0logdir'cron.ERROR'

   Opt.0renamelog = translate(Opt.0renamelog)
   select
      when Opt.0renamelog == '' then
         Opt.0renamelog = 'MONTH'
      when Opt.0renamelog == 'MONTH' then
         nop
      when Opt.0renamelog == 'DAY' then
         nop
      when Opt.0renamelog == 'NEVER' then
         nop
      otherwise do
            logstr = '['date() time()']'
            call lineout Cron.0errorlogfile, '['date() time()'] [ERROR] invalid renamelog value "'Opt.0renamelog'"'
            Opt.0renamelog = 'MONTH'
            call lineout Cron.0errorlogfile, '        using default renamelog = 'Opt.0renamelog
            call lineout Cron.0errorlogfile
         end
   end   /*  end select  */
   call SetLogfileName

   select
      when Opt.0debug == '' then
         Opt.0debug = 0
      when datatype(Opt.0debug, 'number') then
         nop
      otherwise
         call lineout Cron.0errorlogfile, '['date() time()'] [ERROR] invalid debug value "'Opt.0debug'"'
         call lineout Cron.0errorlogfile
         Opt.0debug = 0
   end

   if Opt.0crontab \= '' then
      Cron.0crontab = Opt.0crontab
   else do
         select
            when etc \== '' then
               Cron.0crontab = etc'\crontab'
            when unixroot \== '' then
               Cron.0crontab = unixroot'\etc\crontab'
            otherwise
               Cron.0crontab = ''
         end
      end
   Cron.0crontab = translate(Cron.0crontab, '\', '/')

   if Opt.0pidfile \= '' then
      Cron.0pidfile = Opt.0pidfile
   else do
         select
            when unixroot \== '' then
               Cron.0pidfile = unixroot'\var\run\cron.pid'
            when DirExist(SysBootDrive()'\var\run') == 1 then
               pidfile = SysBootDrive()'\var\run\cron.pid'
            otherwise
               Cron.0pidfile = ''
               call LogIt "[ERROR] variable Cron.0pidfile == ''; there will be no PID file"
         end
      end
   Cron.0pidfile = translate(Cron.0pidfile, '\', '/')

   if Opt.0queue \= '' then
      Cron.0qname = translate(Opt.0queue)
   else
      Cron.0qname = 'CRON_CONTROL'
   return
   /* === End of ResolveOptions ===================================================================== */

   MonthToNum: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  converts month names to numbers                                                                */
   /*  Example: "January,June-August"  returns "1,6-8"                                                */
   /*  called by ReadCrontab;                                                                         */
   /*  this should not really need to know what N is, it is only useful for the error message         */
   /* =============================================================================================== */
   fld = arg(1)
   N = arg(2)
   if length(fld) == 1 then return fld
   /*     word separators are '/', ',', and '-'                                                       */
   /*    for each word, change it to an equivalent number                                             */
   sep = ',/-'
   ThisWrd = ''
   Outstr = ''
   do p = 1 to length(fld) +1
      c = strip(substr(fld, p, 1))
      if c == '' ,
       | pos(c, sep) > 0 then do
            /*  end of word, translate it and add to Out string                                       */
            ThisWrd = translate(ThisWrd)
            if datatype(ThisWrd, 'U') then do
                  select
                     when abbrev(ThisWrd, 'JAN') then
                        ThisWrd = '1'
                     when abbrev(ThisWrd, 'FEB') then
                        ThisWrd = '2'
                     when abbrev(ThisWrd, 'MAR') then
                        ThisWrd = '3'
                     when abbrev(ThisWrd, 'APR') then
                        ThisWrd = '4'
                     when abbrev(ThisWrd, 'MAY') then
                        ThisWrd = '5'
                     when abbrev(ThisWrd, 'JUN') then
                        ThisWrd = '6'
                     when abbrev(ThisWrd, 'JUL') then
                        ThisWrd = '7'
                     when abbrev(ThisWrd, 'AUG') then
                        ThisWrd = '8'
                     when abbrev(ThisWrd, 'SEP') then
                        ThisWrd = '9'
                     when abbrev(ThisWrd, 'OCT') then
                        ThisWrd = '10'
                     when abbrev(ThisWrd, 'NOV') then
                        ThisWrd = '11'
                     when abbrev(ThisWrd, 'DEC') then
                        ThisWrd = '12'
                     otherwise do
                           call LogIt '[ERROR] line 'LineNo()',  invalid month name "'ThisWrd'" in 'Cron.0crontab
                           call LogIt '[ERROR]  "'CrontabEntry(N)'"'
                        end
                  end   /*  end SELECT  */
               end
            Outstr = Outstr||ThisWrd||c
            ThisWrd = ''
         end
      else do
            ThisWrd = ThisWrd||c
         end
   end   /*  end DO  */
   return Outstr
   /* === End of MonthToNum ========================================================================= */

   DayToNum: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    converts day names to numbers                                                                */
   /*    Sunday is day 7, but day 0 will also be converted to 7                                       */
   /*    Examples: "Mon-Fri" returns "1-5"                                                            */
   /*              "Sat,Sun" returns "6,7"                                                            */
   /*              "0,1,2" returns "7,1,2"                                                            */
   /*    a range must be in the right order, lower-upper                                              */
   /*    a range like "Sun-Tue" or "0-2" will return "7-2", which is an invalid range                 */
   /*    if in doubt, list the days instead of using a range--                                        */
   /*    "Sun,Mon,Tue" returns "7,1,2" which is legal                                                 */
   /*  fix: we could allow "7-2" by converting it to "7,1,2" here, but would that be                  */
   /*  compatible with other Crons?                                                                   */
   /* ----------------------------------------------------------------------------------------------- */
   /*  called by ReadCrontab;                                                                         */
   /*  this does not really need to know what N is, but it is useful for the error message            */
   /* =============================================================================================== */
   fld = arg(1)
   N = arg(2)
   /*     word separators are '/', ',', and '-'                                                       */
   /*    for each word, change it to an equivalent number                                             */
   sep = ',/-'
   ThisWrd = ''
   Outstr = ''
   do p = 1 to length(fld) +1
      c = strip(substr(fld, p, 1))
      if c == '' ,
       | pos(c, sep) > 0 then do
            /*  end of word, translate it and add to Out string                                       */
            if ThisWrd == '0' then
               ThisWrd = '7'
            if datatype(ThisWrd, 'M') then do
                  ThisWrd = translate(ThisWrd)
                  select
                     when abbrev(ThisWrd, 'MON') then
                        ThisWrd = '1'
                     when abbrev(ThisWrd, 'TUE') then
                        ThisWrd = '2'
                     when abbrev(ThisWrd, 'WED') then
                        ThisWrd = '3'
                     when abbrev(ThisWrd, 'THU') then
                        ThisWrd = '4'
                     when abbrev(ThisWrd, 'FRI') then
                        ThisWrd = '5'
                     when abbrev(ThisWrd, 'SAT') then
                        ThisWrd = '6'
                     when abbrev(ThisWrd, 'SUN') then
                        ThisWrd = '7'
                     otherwise do
                           call LogIt '[ERROR] line 'LineNo()', invalid day name "'ThisWrd'" in 'Cron.0crontab
                           call LogIt '[ERROR]  "'CrontabEntry(N)'"'
                        end
                  end   /*  end SELECT  */
               end
            Outstr = Outstr||ThisWrd||c
            ThisWrd = ''
         end
      else do
            ThisWrd = ThisWrd||c
         end
   end   /*  end DO  */
   return Outstr
   /* === End of DayToNum =========================================================================== */

   SetLogfileName: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    sets the name of the log file Cron.0logfile                                                  */
   /*    monthly:    "cron-2023-08-Aug.log"                                                           */
   /*    daily       "cron-2023-08-01.log"                                                            */
   /*    never       "cron.log"                                                                       */
   /* =============================================================================================== */
   if Opt.0stdout then
      Cron.0logfile = 'stdout'
   else do
         if Opt.0renamelog == 'NEVER' then do
               Cron.0logfile = Cron.0logdir||'cron.log'
               return Cron.0logfile
            end

         today = date('S')                                                        /* as in "20010101" */
         parse var today yr +4 monNN +2 dayNN .
         today = date('N')                                                     /* as in "01-Jan-2001" */
         parse var today . month .
         select
            when Opt.0renamelog == 'DAY' then
               Cron.0logfile = Cron.0logdir||'cron-'yr'-'monNN'-'dayNN'.log'
            when Opt.0renamelog == 'MONTH' then
               Cron.0logfile = Cron.0logdir||'cron-'yr'-'monNN'-'month'.log'
            when Opt.0renamelog == 'NEVER' then
               Cron.0logfile = Cron.0logdir||'cron.log'
            otherwise do
                  Cron.0logfile = Cron.0logdir||'cron-'yr'-'monNN'-'month'.log'
                  call LogIt '[ERROR] line 'LineNo()', bad value for renamelog "'Opt.0renamelog'", using MONTH'
               end
         end   /*  end select  */
      end
   return Cron.0logfile
   /* === End of SetLogfileName ===================================================================== */

   LogIt: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    Adds a line to the log file.  If an error message, also to the error logfile                 */
   /*    Use this instead of calling lineout directly-- it automatically adds a timestamp             */
   /*    and closes the logfile.                                                                      */
   /* =============================================================================================== */
   /* fix: most messages have a [XXX] prefix, I could use that to decide if and where to send         */
   /*      message. for instance, if Opt.0debug == 'ERROR,RUN' then only messages with those          */
   /*      prefixes would be logged                                                                   */
   parse arg LogMsg
   LogText = '['date() time()']' LogMsg
   call lineout Cron.0logfile, LogText
   call lineout Cron.0logfile
   if word(LogMsg, 1) == '[ERROR]' then do
         call lineout Cron.0errorlogfile, LogText
         call lineout Cron.0errorlogfile
      end
   return
   /* === End of LogIt ============================================================================== */

   LogError: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  This is only called by the error handlers.                                                     */
   /*  It sends error message both to the logfile and to the error file.                              */
   /*  fix: I should probably merge this with Logit()                                                 */
   /*                                                                                                 */
   /*  The error handlers do not do much.  A syntax problem is fatal; otherwise try                   */
   /*  to keep going, on the grounds that a partly working cron daemon is better than                 */
   /*  none.  Error messages are also written to a separate error log to make them                    */
   /*  slightly more noticeable to the user.                                                          */
   /*                                                                                                 */
   /*  fix: Of course, it would be better to mail error messages.                                     */
   /* =============================================================================================== */
   /* if I cannot even write to a log, give up!                                                       */
   signal on NotReady name Crash
   parse arg ehRC ehSIGL msg
   /* RC must be a number from 0 to 60 or errortext() complains                                       */
   if ehRC == 'RC' then ehRC = 0

   call lineout Cron.0logfile, '['date() time()'] [ERROR]' msg
   call lineout Cron.0logfile, '                               ' G.0me': line 'ehSIGL 'RC= 'ehRC
   call lineout Cron.0logfile, '                               ==> 'sourceline(ehSIGL)
   if ehRC \== 0 then
      call lineout Cron.0logfile, '                               ' '('errortext(ehRC)')'
   call lineout Cron.0logfile

   /*  if the two log files are both 'stdout', no need to repeat it                                   */
   if Cron.0errorlogfile \= Cron.0logfile then do
         call lineout Cron.0errorlogfile, '['date() time()'] [ERROR]' msg
         call lineout Cron.0errorlogfile, '                               ' G.0me': line 'ehSIGL 'RC= 'ehRC
         call lineout Cron.0errorlogfile, '                               ==> 'sourceline(ehSIGL)
         if ehRC \== 0 then
            call lineout Cron.0errorlogfile, '                               ' '('errortext(ehRC)')'
         call lineout Cron.0errorlogfile
      end
   call on Notready
   return
   /* === End of LogError =========================================================================== */

   Crash: /*fold00*/
   /* =============================================================================================== */
   /*  this is the error handler for LogError                                                         */
   /*    something happened that was so bad we cannot even log the error                              */
   /* =============================================================================================== */
   call lineout 'stderr', 'Fatal Error: rc='rc' sigl='sigl' ehRC='ehRC' ehSIGL='ehSIGL
   call Egress
   /* === End of Crash ============================================================================== */

   Syntax: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   Esigl = sigl
   call LogError rc Esigl '[Syntax handler]'
   call Egress
   return
   /* === End of Syntax ============================================================================= */

   Error: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   Esigl = sigl
   call LogError rc Esigl '[Error handler]'
   return
   /* === End of Error ============================================================================== */

   Failure: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   Esigl = sigl
   call LogError rc Esigl '[Failure handler]'
   return
   /* === End of Failure ============================================================================ */

   Notready: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   Esigl = sigl
   call LogError rc Esigl '[NotReady handler] at line' Esigl
   return
   /* === End of Notready =========================================================================== */

   NoValue:
   /* =============================================================================================== */
   /* =============================================================================================== */
   Esigl = sigl
   call LogIt '[ERROR] NoValue at line' Esigl' for variable "'condition('d')'"'
   call LogIt '        ==> 'sourceline(Esigl)
   call Egress
   /* === End of NoValue ============================================================================ */

   Egress: /*fold00*/
   /* =============================================================================================== */
   /*    cleanup before exiting. All exits should go through here                                     */
   /* =============================================================================================== */
   call LogIt '[Egress] delete PID file, delete Queue, exit'
   call SysFileDelete Cron.0pidfile
   call rxqueue 'delete', Cron.0qname
   call LogIt '==============================================================================='
   exit 0
   /* === End of Egress ============================================================================= */

   Help: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   call lineout 'stderr', 'Usage:'
   call lineout 'stderr', '    'G.0me '[option] [option] ...'
   call lineout 'stderr', 'Options are:'
   call lineout 'stderr', '    -debug=<level>                  (default 0)'
   call lineout 'stderr', '    -crontab=<file name>            (default: %etc\crontab or %unixroot\etc\crontab)'
   call lineout 'stderr', '    -pidfile=<filename>             (default: %unixroot\var\run\cron.pid)'
   call lineout 'stderr', '    -queue=<name of control queue>  (default: CRON_CONTROL)'
   call lineout 'stderr', '    -logdir=<log directory>         (default: %unixroot\var\log\)'
   call lineout 'stderr', '    -renamelog=NEVER|MONTH|DAY      (default: MONTH)'
   call lineout 'stderr', '                                    NEVER == <logdir>\cron.log'
   call lineout 'stderr', '                                    MONTH == new log file every month'
   call lineout 'stderr', '                                    DAY   == new log file every day'
   call lineout 'stderr', '    -stdout                         send messages to STDOUT instead of log file'
   call lineout 'stderr', ''
   call lineout 'stderr', ''
   call lineout 'stderr', 'control commands that can be sent through the control queue include'
   call lineout 'stderr', ' STOP, PAUSE, RESUME, DEBUG'
   call lineout 'stderr', 'See "StartCron.cmd HELP" for more information about them.'
   call lineout 'stderr', ''
   call lineout 'stderr', 'Current configuration is:'
   call lineout 'stderr', '    logfile     == 'Cron.0logfile
   call lineout 'stderr', '    rename log  == 'Opt.0renamelog
   call lineout 'stderr', '    debug level == 'Opt.0debug
   call lineout 'stderr', '    crontab     == 'Cron.0crontab
   if \FileExist(Cron.0crontab) then
      call lineout 'stderr', '                   (but it does not exist)'
   call lineout 'stderr', '    pidfile     == 'Cron.0pidfile
   call lineout 'stderr', '    queue       == 'Cron.0qname
   call lineout 'stderr', ''
   call lineout 'stderr', 'However,' G.0me 'is usually started from StartCron.cmd, so'
   call lineout 'stderr', '      StartCron.cmd HELP'
   call lineout 'stderr', 'is probably what you want'
   return 1
   /* === End of Help =============================================================================== */

   LineNo: /*fold00*/
   0LineNo:
   /* =============================================================================================== */
   /*  returns the number of the current line--    say 'now at line 'LineNo()                         */
   /* =============================================================================================== */
   _ = sigl
   return _
   /* === End of 0LineNo ============================================================================ */

   Mark: /*fold00*/
   0Mark:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       CALL Mark [<list of variable names>] [, <string>]                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Shows the filename and line number it was called from, the current values of the               */
   /*  variables named, and an informational string                                                   */
   /*  Arg(1) is a list of variable names                                                             */
   /*  Arg(2) can be any descriptive text                                                             */
   /*  Examples:                                                                                      */
   /*      CALL mark 'varA varB', 'At top of the "DO x = 1 to 5" loop'                                */
   /*  If no variable names are given, it will just show a message:                                   */
   /*      CALL mark '', 'At top of the "DO x = 1 to 5" loop'                                         */
   /*                                                                                                 */
   /*  This must share all the variables of whatever routine it was called from.                      */
   /*  I cannot think of a way to do it without setting at least one variable here,                   */
   /*  so reserve the variable '_' for the use of Mark and similar routines.                          */
   /* =============================================================================================== */
   _ = sigl
   call lineout 'stderr', ''
   if arg(2, 'e') ,
    & arg(2) = '' then nop
   else call lineout 'stderr', '[Mark]['value('G.0me')':'_']' arg(2)
   do _ = 1 to 99
      if word(arg(1), _) = '' then leave _
      if symbol(word(arg(1), _)) = 'VAR' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||'= "'||value(word(arg(1), _))||'"'
      if symbol(word(arg(1), _)) = 'LIT' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not defined'
      if symbol(word(arg(1), _)) = 'BAD' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not a variable'
   end _
   drop _
   return 0
   /* === End of 0Mark ============================================================================== */

   /***************************************************************************************************/
   /*  start what were external procedures                                                            */
   /***************************************************************************************************/

   GetOpts: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     INTERPRET GetOpts(args [,valid])                                                            */
   /*                                                                                                 */
   /*  where 'args' is the command line to be parsed, and                                             */
   /*  'valid' is an optional space-separated list of valid switches                                  */
   /*  'help' and '?' are always valid                                                                */
   /*                                                                                                 */
   /*  Example:                                                                                       */
   /*     parse arg args                                                                              */
   /*     interpret GetOpts(args, 'switchA switchB')                                                  */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Switches must start with '-'. For example:                                                     */
   /*                                                                                                 */
   /*         -Foo                (Opt.0Foo == 1)                                                     */
   /*         -Foo=1              (Opt.0Foo == 1)                                                     */
   /*         -!Foo or -Foo=0     (Opt.0Foo == 0)                                                     */
   /*         -Foo=string         (Opt.0Foo == 'string')                                              */
   /*         -Foo="string"       (Opt.0Foo == 'string')                                              */
   /*         -Foo=a,b,c          (Opt.0Foo == 'a,b,c')                                               */
   /*         -Foo="a, b, c"      (Opt.0Foo == 'a, b, c')                                             */
   /*         -Foo="a b c"        (Opt.0Foo == 'a b c')                                               */
   /*         -Foo=123            (Opt.0Foo == '123')                                                 */
   /*         -Foo= or -Foo=''    (Opt.0Foo == '')                                                    */
   /*                                                                                                 */
   /*  switches must not have spaces before or after the '='                                          */
   /*  values with spaces must be in quotes. the quotes will be stripped.                             */
   /*                                                                                                 */
   /*  a switch without a value is set to '1'                                                         */
   /*        -switchA                                                                                 */
   /*  is the same as                                                                                 */
   /*        -switchA=1                                                                               */
   /*                                                                                                 */
   /*  a switch can be negated with "!", so                                                           */
   /*        -!switchA                                                                                */
   /*  is the same as                                                                                 */
   /*        -switchA=0                                                                               */
   /*                                                                                                 */
   /*     -? is the equivalent of -help                                                               */
   /*     they will both result in:                                                                   */
   /*        Opt.0help    == 1                                                                        */
   /*                                                                                                 */
   /*  Interpreting what GetOpts returns will create several variables that the                       */
   /*  caller can use--                                                                               */
   /*     Opt.              is a stem variable that holds whatever was specified by switches          */
   /*     Opt.0optlist      a space-separated list of options that were read                          */
   /*     Opt.0errors       number of errors; that is, when GetOpts() did not understand something    */
   /*                                                                                                 */
   /*                       elements of the Opt. compound variable all start with '0', to avoid       */
   /*                       name conflicts with simple variables                                      */
   /*                                                                                                 */
   /*     ArgV.             is a stem that holds the non-switch arguments                             */
   /*                       (ArgV.0 is the name of the .cmd script, NOT the number of items           */
   /*                       in the stem)                                                              */
   /*     ArgC              is the number of non-switch arguments                                     */
   /*                                                                                                 */
   /*  Example:                                                                                       */
   /*     Foo.cmd -dir=up -msg="that way" -v "From here"                                              */
   /*     will result in these variables:                                                             */
   /*        Opt.0optlist == opt.0dir opt.0msg                                                        */
   /*        Opt.0dir     == up                                                                       */
   /*        Opt.0msg     == that way                                                                 */
   /*        Opt.0v       == 1                                                                        */
   /*        ArgV.0       == Foo.cmd                                                                  */
   /*        ArgV.1       == From here                                                                */
   /*        ArgC         == 1                                                                        */
   /*                                                                                                 */
   /*  in the case of errors, GetOpts() will set Opt.0errors to the number of errors                  */
   /*  found, and cause the caller to show a message about it on stderr. The caller                   */
   /*  should probably abort if Opt.0errors is greater than 0.                                        */
   /* =============================================================================================== */
   /* Some important variables (inside GetOpts) are:                                                  */
   /*    Str.SetOpts       == commands the the caller will execute to set options                     */
   /*                          in the Opt. stem variable                                              */
   /*    Str.SetArgs       == commands the the caller will execute to put the arguments               */
   /*                          in the ArgV. and ArgC variables                                        */
   /*    Str.Errors        == commands the caller will execute to report errors                       */
   /*                          reading the command line                                               */
   /*    ReturnStr         == all of the above, combined and returned to caller                       */
   /*    flag.             == variables for GetOpt's internal use                                     */
   /*                          flag.error == number of errors trying to read command line             */
   /*    Str.SwitchList    == list of options (switches) used on command line                         */
   /*    ArgC              == number of non-switch arguments                                          */
   /*    ArgV.             == array of non-switch arguments                                           */
   /*    valid             == list of valid switches                                                  */
   /*                                                                                                 */
   /* =============================================================================================== */
   flag. = ''
   flag.errors = 0
   str. = ''
   ReturnStr = ''
   ArgC = 0
   ArgV. = ''
   parse arg args , valid
   do i = 1 to WordsQ(args)
      thisArg = WordQ(args, i)
      if left(thisArg, 1) == '-' then
         call ReadSwitch thisArg
      else do
            /*    not a switch, add to Arg list                                                       */
            ArgC = ArgC +1
            ArgV.ArgC = thisArg
         end
   end
   /* ----------------------------------------------------------------------------------------------- */
   /*    now put it all together into a series of commands for the caller to interpret                */
   if flag.errors > 0 then do
         str.SwitchList = strip(str.SwitchList' Opt.0Errors')
      end
   str.SetOpts = str.SetOpts' Opt.0Errors = 'flag.errors ';'
   str.SetOpts = str.SetOpts 'Opt.0optlist = "'str.SwitchList'" ;'

   str.SetArgs = 'ArgV. = "" ; parse source . . ArgV.0 ;  ArgC = "'ArgC'" ;'
   do i = 1 to ArgC
      varVal = Unquoted(ArgV.i)
      str.SetArgs = str.SetArgs' ArgV.'i' = 'Quoted(varVal)' ;'
   end

   ReturnStr = str.errors str.SetOpts str.SetArgs
   return ReturnStr
   /* === End of GetOpts ============================================================================ */

   ReadSwitch: procedure expose flag. str. valid (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  this is part of GetOpts()                                                                      */
   /* =============================================================================================== */
   parse arg thisArg
   Negate = 0
   switch = strip(thisArg, 'L', '-')
   if left(switch, 1) == '!' then do
         switch = strip(switch, 'L', '!')
         Negate = 1
      end
   parse var switch optName '=' optVal
   select
      /* -? and -help have fixed meanings and are allowed even if not included in 'valid'             */
      when translate(optName) == '?' then
         optName  = 'Opt.0Help'
      when translate(optName) == 'HELP' then
         optName  = 'Opt.0Help'
      when optName == '' then do
            str.errors = ,
              str.errors||'call lineout "stderr", "[GetOpts] `-` without a switch name" ;'
            flag.errors = flag.errors +1
            return
         end
      when valid \= '' ,
         & wordpos(translate(optName), translate(valid)) == 0 then do
            str.errors = str.errors|| ,
                         'call lineout "stderr", "[GetOpts] I do not understand switch `'thisArg'`" ;'
            flag.errors = flag.errors +1
            return
         end
      otherwise
         optName = 'Opt.0'optName
   end   /*  select  */

   select
      when Negate then
         if optVal == '' then
            optVal = 0
         else do
               str.errors = str.errors|| ,
                            'call lineout "stderr", "[GetOpts] I do not understand switch `'thisArg'`" ;'
               flag.errors = flag.errors +1
               return
            end
      when optVal == '' ,
         & right(switch, 1) \= '=' then
         /*  "-switch=" means Opt.0switch == '', not 1                                                */
         optVal = '1'
      otherwise
         /*    try to be tolerant of unusual arguments; only strip quotes if they are balanced        */
         /*    and only one level of nested quotes                                                    */
         optVal = Unquoted(optVal)
   end   /*  select  */

   if wordpos(optName, str.SetOpts) > 0 then do
         str.errors = ,
           str.errors||'call lineout "stderr", "[GetOpts] duplicate switch `'thisArg'`" ;'
         flag.errors = flag.errors +1
      end
   else do
         str.SetOpts = str.SetOpts' 'optName' = 'Quoted(optVal)' ;'
         str.SwitchList = strip(str.SwitchList' 'optName)
      end
   return
   /* === End of ReadSwitch ========================================================================= */

   FileExist: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        if FileExist(<filename>) then                                                            */
   /*                                                                                                 */
   /*  returns 1 (True) if <filename> exists and 0 (False) if it does not.                            */
   /*                                                                                                 */
   /*  <filename> must be a real file. This returns 0 for devices or for streams like STDOUT.         */
   /*  Granted, STDOUT does exist, but it is more useful this way. I mostly use this when I want      */
   /*  to avoid clobbering an existing file,                                                          */
   /*                                                                                                 */
   /*  The main reason for this is that                                                               */
   /*        if FileExist(filename) then                                                              */
   /*  is easier to remember than                                                                     */
   /*         if stream(filename, 'c', 'query exists') then                                           */
   /* =============================================================================================== */
   /*  Most of the complications are because OS/2 Rexx and Regina return different values             */
   /*  from stream() for stdout and for devices. Regina usually returns 0 for stdout, but             */
   /*  not always. I don't know why.                                                                  */
   parse arg FileN
   if FileN = '' then return 0
   if wordpos(translate(FileN), 'STDOUT STDERR') > 0 then return 0
   ret = stream(FileN, 'c', 'query exists')
   if ret == '' then return 0
   if left(translate(ret), 5) == '\DEV\' then return 0
   if left(translate(ret), 5) == '/DEV/' then return 0
   return 1
   /* === End of FileExist ========================================================================== */

   WordQ: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordQ(<string>, <word number>, <separator characters>, <quote characters>)            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a replacement for the Rexx Word() function.                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Word() function--                                                                     */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg str , thisnum , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0
   Thisword  = ''
   w1        = 0
   w2        = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(str) ,
          & pos(substr(str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(str) then numwords = numwords +1
   if numwords == thisnum then w1 = i

   do while (i <= length(str))
      if pos(substr(str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(str) ,
                   & substr(str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(str, i, 1), chrsW) \= 0 ,
       & i <= length(str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            /* unless the end of the word we want, then quit                                          */
            if w1 > 0 then do
                  w2 = i -1
                  leave
               end
            do while (pos(substr(str, i, 1), chrsW) \= 0) ,
                   & (i <= length(str))
               i = i +1
            end
            if i <= length(str) then numwords = numwords +1
            /* if this is the start of the word we want, note position in string                      */
            if numwords == thisnum then w1 = i
         end   /*  wordsep  */
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   if i > length(str) then i = length(str)
   if w1 > 0 then do
         /* we know w1, if we ran out of string, w2 is the end of string                              */
         if w2 == 0 then w2 = i
         /* found start and end of word, copy it                                                      */
         wlen = w2 - w1 +1
         Thisword = substr(str, w1, wlen)
      end
   return Thisword
   /* === End of WordQ ============================================================================== */

   WordsQ: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordsQ(string, <separator characters>, <quote characters>)                            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a drop-in replacement for the Rexx Word() function.                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Words() function--                                                                    */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg str , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(str) ,
          & pos(substr(str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(str) then numwords = numwords +1

   do while (i <= length(str))
      if pos(substr(str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(str) ,
                   & substr(str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(str, i, 1), chrsW) \= 0 ,
       & i <= length(str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            do while (pos(substr(str, i, 1), chrsW) \= 0) ,
                   & (i <= length(str))
               i = i +1
            end
            if i <= length(str) then numwords = numwords +1
         end
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   return numwords
   /* === End of WordsQ ============================================================================= */

   Quoted: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Quoted(string)                                                                  */
   /*                                                                                                 */
   /*    if possible, quotes a string with the kind of quotes it does not contain                     */
   /*        {He said "no"}   --> {'He said "no"'}                                                    */
   /*        {He said 'no'}   --> {"He said 'no'"}                                                    */
   /*    otherwise uses double quotes and lets the caller deal with the errors                        */
   /* =============================================================================================== */
   parse arg string
   select
      when pos('"', string) == 0 then
         newstring = '"'string'"'
      when pos("'", string) == 0 then
         newstring = "'"string"'"
      otherwise
         newstring = '"'string'"'
   end
   return newstring
   /* === End of Quoted ============================================================================= */

   Unquoted: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Unquoted(string)                                                                */
   /*                                                                                                 */
   /*    strips quotes from a string if and only if it is enclosed in a balanced pair                 */
   /*    of single or double quotes                                                                   */
   /* =============================================================================================== */
   parse arg string
   if length(string) < 2 then return string
   select
      when left(string, 1) == '"' ,
         & right(string, 1) == '"' then
         string = substr(string, 2, length(string) -2)
      when left(string, 1) == "'" ,
         & right(string, 1) == "'" then
         string = substr(string, 2, length(string) -2)
      otherwise
         nop
   end
   return string
   /* === End of Unquoted =========================================================================== */

   DirExist: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*      DirExist(<directory name>)                                                                 */
   /*                                                                                                 */
   /*  Examples:                                                                                      */
   /*      if DirExist('F:\os2') then ...                                                             */
   /*      if \DirExist('F:\NotThere') then ...                                                       */
   /*      if \DirExist(filespec('path', outfile)) then ...                                           */
   /*                                                                                                 */
   /*  returns 1 if a directory exists, 0 if it does not                                              */
   /*                                                                                                 */
   /*  use either \ or /                                                                              */
   /*  trailing slashes are stripped                                                                  */
   /*      'F:\OS2' or 'F:\OS2\'   both mean 'F:\OS2'                                                 */
   /*      'F:' or 'F:\'           both mean 'F:\'                                                    */
   /*                                                                                                 */
   /*  in 4os2.exe/cmd.exe,                                                                           */
   /*      'F:'                    means "the current directory on drive F:"                          */
   /*  but that would be pointless here; the current directory must exist                             */
   /*  likewise for                                                                                   */
   /*      "DirExist('.')"                                                                            */
   /*                                                                                                 */
   /*  if no directory name is provided at all, it is taken as '.'                                    */
   /*  that way a calling routine can verify the path of a filename without the                       */
   /*  extra step of checking the output of filespec() for ''.                                        */
   /*          if \DirExist(filespec('path', outfile)) then do                                        */
   /*              say 'invalid file name, no such directory'                                         */
   /*              return 1                                                                           */
   /*          end                                                                                    */
   /*                                                                                                 */
   /*  relative paths like "..\..\bin" with leading ".." are legal, but                               */
   /*  SysFileTree() can easily be confused by them                                                   */
   /*                                                                                                 */
   /*  wildcards are legal in <directory name>, but probably not useful                               */
   /* =============================================================================================== */

   /* ----------------------------------------------------------------------------------------------- */
   /*  the root directory of a drive needs special handling because SysFileTree()                     */
   /*  will not find it unless it has at least one subdirectory.                                      */
   /*                                                                                                 */
   /*  There is a logic to that--                                                                     */
   /*  to SysFileTree(), "F:\OS2" means "the directory named F:\OS2".                                 */
   /*  An ending slash, "F:\OS2\" means "all subdirectories of F:\OS2".                               */
   /*  so there is no way to refer to the root directory of a drive.                                  */
   /*  "R:\" means "all subdirectories under R:\", and                                                */
   /*  "R:" means "the current directory on R:".                                                      */
   /* ----------------------------------------------------------------------------------------------- */
   parse arg DirName .
   if DirName == '' then return 1
   if DirName == '.' then return 1
   DirName = translate(DirName, '\', '/')

   /*  if DirName = 'F:\' verify that drive F: exists. If so, it presumably has a root directory      */
   if right(DirName, 2) == ':\' then do
         DriveN = filespec('drive', DirName)
         if SysDriveInfo(DriveN) = '' then
            return 0
         else
            return 1
      end
   else do
         DirName = strip(DirName, 't', '\')
         ret = SysFileTree(DirName, 'Dirlist.', 'do')
         if Dirlist.0 > 0 then
            return 1
         else
            return 0
      end
   /* === End of DirExist =========================================================================== */

