/**/

   /***************************************************************************************************/
   /*  STARTCRON.CMD to start/stop/pause/resume the Cron.cmd daemon                                   */
   /*  Anton Monroe <amonroe5301@gmail.com>                                                           */
   /*                                                                                                 */
   /*  If you want to work on StartCron.cmd or Cron.cmd, you might wonder about a                     */
   /*  couple of things: The "fold00" comments are for the folding text editor I use.                 */
   /*  If you use FTE or eFTE you should see my folds. The routine names that start                   */
   /*  with "0" are part of my own scheme for maintaining commonly-used routines in my                */
   /*  scripts. (See RexxTool.zip at hobbesarchive.com). The Mark routine is only used                */
   /*  for debugging; it will show the value of the given variable(s) at any point.                   */
   /*                                                                                                 */
   /***************************************************************************************************/
   /* Load RexxUtil and RXU libraries                                                                 */
   if RxFuncQuery('SysLoadFuncs') then do
         call RxFuncAdd 'SysLoadFuncs', 'RexxUtil', 'SysLoadFuncs'
         call SysLoadFuncs
      end
   if RxFuncQuery('rxqprocstatus') then
      call RxFuncAdd 'RxuInit', 'RXU', 'RxuInit'
   call RxuInit

   Setup: /*fold00*/
   /* =============================================================================================== */
   /*  variables in this section are the only things you might need to change.                        */
   /*  if a variable is unset, Cron.cmd will use the defaults.                                        */
   /* =============================================================================================== */
   unixroot = value('UNIXROOT', , 'OS2ENVIRONMENT')
   /*  This assumes Cron.cmd is in the same directory as StartCron.cmd                                */
   parse source . . me
   thisdir = filespec('drive', me)||filespec('path', me)
   CronName = thisdir||'Cron.cmd'
   if \FileExist(CronName) then do
         call lineout 'stderr', 'cannot find' CronName
         return 1
      end
   drop me
   /*  Cron.cmd can determine defaults for these things, but StartCron.cmd                            */
   /*  also needs to know what some of them are, so set them here and pass them on                    */
   /*  the Cron.cmd command line.                                                                     */
   /*                                                                                                 */
   /*  the name of the queue that StartCron.cmd uses to send commands to the detached Cron.cmd        */
   Qname = 'CRON_CONTROL'
   /*                                                                                                 */
   /*  how often to rename log file; choices are 'month', 'day', or 'never'                           */
   /*  default is 'month'                                                                             */
   renamelog = 'month'
   /*                                                                                                 */
   /*  The starting debug level for cron.cmd. 0 is normal; 3 is more detail than you                  */
   /*  probably want. You can also change it while Cron.cmd is running.                               */
   /*  You can also change this from the command line with "StartCron.cmd START -debug=3"             */
   debuglevel = 0
   /*                                                                                                 */
   if unixroot == '' then do
         /*  if there is no unixroot, you probably have no directories like /var/log or               */
         /*  /var/run. Put everything in the directory of StartCron.cmd                               */
         pidfile = thisdir||'cron.pid'
         logdir = thisdir
         crontab  = thisdir||'crontab'
      end
   else do
         /*  pidfile is the name of the PID file. default is %unixroot\var\run\cron.pid               */
         /*  but I have most of my init scripts use \var\run\init\ so I can remove                    */
         /*  old pid files more easily                                                                */
         /*  if \var\run\init\ exists, use it instead of default                                      */
         if DirExist(unixroot||'\var\run\init') then
            pidfile = unixroot||'\var\run\init\cron.pid'
         else
            pidfile = unixroot||'\var\run\cron.pid'
         /*                                                                                           */
         /*  where to keep log file(s)                                                                */
         /*  I prefer a separate directory for the cron logs. If it exists, use it.                   */
         if DirExist(unixroot||'\var\log\cron') then
            logdir = unixroot||'\var\log\cron\'
         else
            logdir = unixroot||'\var\log\'
         /*                                                                                           */
         /*  the crontab file. default is %etc\crontab or %unixroot\etc\crontab                       */
         etc = value('ETC', , 'OS2ENVIRONMENT')
         if etc \= '' then
            crontab = etc||'\crontab'
         else
            crontab = unixroot||'\etc\crontab'
      end   /*  if unixroot  */
   /* =============================================================================================== */

   parse arg args
   Opt. = ''
   interpret GetOpts(args)

   /*  a quick hack to let me set the starting debug value from the StartCron.cmd command line:       */
   /*  fix: I could add other options in the same way                                                 */
   if Opt.0debug \= '' then
      debuglevel = Opt.0debug

   /*  if any of these variables were set in the block above, pass them on as                         */
   /*  switches to Cron.cmd                                                                           */
   CronCmd = CronName
   if symbol('debuglevel') == 'VAR' then
      CronCmd = CronCmd '-debug='debuglevel
   if symbol('crontab') == 'VAR' then
      CronCmd = CronCmd '-crontab='crontab
   if symbol('logdir') == 'VAR' then
      CronCmd = CronCmd '-logdir='logdir
   if symbol('renamelog') == 'VAR' then
      CronCmd = CronCmd '-renamelog='renamelog
   if symbol('pidfile') == 'VAR' then
      CronCmd = CronCmd '-pidfile='pidfile
   if symbol('Qname') == 'VAR' then
      CronCmd = CronCmd '-queue='Qname
   /*                                                                                                 */
   /*  the command that will detach Cron.cmd                                                          */
   DetachCronCmd = '@detach 'CronCmd
   /*  StartCronCmd can run Cron.cmd in a VIO window instead of detached. It might be                 */
   /*  useful for testing. Edit as needed.  Use "StartCron.cmd START -foreground"                     */
   /*  The -stdout switch sends all logging to screen                                                 */
   StartCronCmd = '@start "Cron Daemon" /i /f 'CronCmd '-stdout'
   /*                                                                                                 */
   /*  I prefer to run it under 4os2 because it gives me a bigger window-- the log messages           */
   /*  need a wide screen                                                                             */
   if ShellName() == '4OS2' then
      StartCronCmd = '@start /fg "Cron Daemon" 4os2.exe /c 'CronCmd '-stdout'
   /* === End of Setup ============================================================================== */

   Main: /*fold00*/
   /*=================================================================================================*/
   /*  set some variables, then read the command line and do something                                */
   /*=================================================================================================*/
   G. = ''
   /*  G.0me' is the name of this file. (When a Rexx script is invoked from the                       */
   /*  command line, the name is in ALL CAPS, which is hard to read.  To get the                      */
   /*  correct capitalization of it, the only way I have found is a cumbersome call to                */
   /*  SysFileTree(), and even that only corrects the filename, not the path.)                        */
   parse source . . G.0me
   call SysFileTree G.0me, 'tmp.', 'FO'
   G.0me_long = tmp.1
   G.0me = filespec('name', G.0me_long)
   G.0meID = '['G.0me']'

   /*    the plan, someday, is to return some rational exit codes:                                    */
   Err.0none = 0
   Err.0usage = 1
   Err.0status = 2

   /*    list of shared variables                                                                     */
   Shared = 'G. Opt. Qname CronCmd CronName pidfile Err.'
   signal on NoValue

   pid = ''
   IsQ = 0

   if Opt.0help \= '' then do
         call Help
         return Err.0usage
      end
   /* ----------------------------------------------------------------------------------------------- */

   if \FileExist(CronName) then do
         say 'Cannot find the Cron program' CronName
         return Err.0usage
      end

   Cmd = translate(ArgV.1)
   /*  call GetStatus to set value of variables pid and IsQ                                           */
   call GetStatus
   select
      when Cmd == 'START' then do
            if IsQ then
               say CronName 'is already running (ie, the queue 'Qname' exists)'
            else do
                  if \FileExist(crontab) then do
                        say 'Cannot find the crontab file' crontab
                        return Err.0usage
                     end
                  if Opt.0foreground == 1 then
                     address cmd StartCronCmd
                  else
                     address cmd DetachCronCmd
                  do c = 1 to 8
                     if QueueExist(Qname) then leave c
                     call SysSleep 1
                  end c
                  call ShowStatus
               end
         end
      when Cmd == 'STOP' then do
            if \IsQ then
               say CronName 'is not running'
            else
               call SendCmd 'STOP'
         end
      when Cmd == 'PAUSE' then do
            if \IsQ then
               say CronName 'is not running'
            else do
                  if datatype(ArgV.2, 'whole') then
                     call SendCmd 'PAUSE 'ArgV.2
                  else
                     say 'invalid PAUSE value: 'ArgV.2
               end
         end
      when Cmd == 'RESUME' then do
            if \IsQ then
               say CronName 'is not running'
            else
               call SendCmd 'RESUME'
         end
      when Cmd == 'DEBUG' then do
            if \IsQ then
               say CronName 'is not running'
            else do
                  if datatype(ArgV.2, 'whole') then
                     call SendCmd 'DEBUG 'ArgV.2
                  else
                     say 'invalid DEBUG value: 'ArgV.2
               end
         end
      when Cmd == 'STATUS' then
         call ShowStatus
      when Cmd == 'HELP' then do
            call Help
            return Err.0usage
         end
      when Cmd == 'KILL' then do
            call Kill
            call ShowStatus
         end
      when Cmd == '' then do
            call Help
            return Err.0usage
         end
      otherwise do
            say 'unknown command: "'ArgV.1'"'
            say ''
            call Help
            return Err.0usage
         end
   end
   return Err.0none
   /* === End of Main =============================================================================== */

   SendCmd: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    sends a command to the queue                                                                 */
   /* =============================================================================================== */
   /*  fix: If I send cron.cmd a command from the control program, there will be up to                */
   /*        one minute delay before cron.cmd reads the queue and StartCron.cmd can                   */
   /*        verify that it happened.                                                                 */
   /*        There may be a way to avoid that but I do not know what it is.                           */
   /* ----------------------------------------------------------------------------------------------- */
   parse arg Qcmd
   /*    switch to Cron queue                                                                         */
   oldQname = rxqueue('set', Qname)
   queue Qcmd
   timestr = time()
   parse var timestr hh ':' mm ':' ss
   towait = 60 - ss
   say 'sent 'Qcmd' to queue 'Qname', it should take effect in 'towait' seconds'
   key = ''
   say 'waiting for Cron.cmd to read the queue.'
   say 'Press any key to return to the command line now'
   do c = 1 to (towait +10)
      if queued() == 0 then leave
      if chars() == 0 then do
            call charout 'stdout', '.'
            call SysSleep 1
         end
      else do
            key = SysGetKey('noecho')
            leave
         end
   end
   if key == '' then do
         if queued() > 0 then do
               say ''
               say G.0meID' Cron.cmd should have read the queue by now, but it still has 'queued()' lines'
               say '        I am giving up. Try "StartCron.cmd status"'
            end
      end

   /*    switch back to session queue                                                                 */
   call rxqueue 'set', oldQname
   return 0
   /* === End of SendCmd ============================================================================ */

   GetStatus: /*fold00*/
   /* =============================================================================================== */
   /*  checks status of queue and pid file                                                            */
   /*  exposes pid of Cron                                                                            */
   /*  exposes IsQ == 1 if the control queue exists                                                   */
   /* =============================================================================================== */
   Esigl = sigl
   IsQ = QueueExist(Qname)
   pid = ''
   if FileExist(pidfile) then do
         pid = linein(pidfile, 1)
         call stream pidfile, 'c', 'CLOSE'
      end
   if (IsQ == 1 ,
    & pid == '') ,
    | (IsQ == 0 ,
    & pid \= '') then do
         say '[Error:] conflict between queue and PID; try' G.0me 'status'
         call Mark 'Esigl Qname IsQ pid pidfile'
      end
   return
   /* === End of GetStatus ========================================================================== */

   ShowStatus: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    show if the Cron daemon is running, and some information about it                            */
   /* =============================================================================================== */
   call GetStatus
   if FileExist(pidfile) then do
         StartTime = SysGetFileDateTime(pidfile, 'create')
         ModTime = SysGetFileDateTime(pidfile, 'write')
         pid = linein(pidfile, 1)
         CronParam = linein(pidfile)
         call stream pidfile, 'c', 'CLOSE'
      end

   select
      when IsQ == 1 ,
         & pid \= '' then do
            say CronName 'has PID' pid
            say 'It was started at     'right(DayOfWeek(StartTime), 10) StartTime
            say 'State last changed at 'right(DayOfWeek(ModTime), 10) ModTime
            say 'Its current state is:'
            do while CronParam \= ''
               parse var CronParam ParN '=' ParV CronParam
               say '      'left(ParN, 14, ' ')'== 'ParV
            end
            say '      there are 'RxQueued(Qname) 'line(s) in the queue'
         end
      when IsQ == 0 ,
         & pid == '' then
         say CronName 'is not running'
      otherwise
         say 'There is conflicting information:'
         if IsQ == 1 then
            say '    the control queue exists'
         else
            say '    the control queue does not exist'
         if FileExist(pidfile) then
            say '    the file' pidfile 'exists'
         else
            say '    the file' pidfile 'does not exist'
         if pid \== '' then do
               say '    the PID is' pid
               if IsAlive(pid) then
                  say '    the process with that PID is running'
               else
                  say '    but there is no process running with that PID'
            end
         else
            say '    there is no PID'
         say 'You could try'
         say '    'G.0me 'STOP  or'
         say '    'G.0me 'KILL'
   end   /*  select  */
   return
   /* === End of ShowStatus ========================================================================= */

   QueueExist: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*   Usage:    if QueueExist(QueueName) then ...                                                   */
   /*                                                                                                 */
   /*   returns 1 if the control queue exists, 0 if it does not                                       */
   /* =============================================================================================== */
   parse arg Qname

   /* if Qname is current, return True                                                                */
   if Qname == rxqueue('get') then
      return 1
   else do
         /*    try to create Qname; if it already exists, rxqueue() will create an                    */
         /*    alternate queue and return its name                                                    */
         /*    if Qname already existed, delete the alternate and return 1 (True)                     */
         /*    if Qname did not already exist, delete Qname and return 0 (False)                      */
         newQ = rxqueue('CREATE', Qname)
         if newQ \= Qname then do
               call rxqueue 'delete', newQ
               return 1
            end
         if newQ == Qname then do
               call rxqueue 'delete', Qname
               return 0
            end
      end
   return 0
   /* === End of QueueExist ========================================================================= */

   Kill: /*fold00*/
   /* =============================================================================================== */
   /*  Removes the PID file and the control queue.                                                    */
   /*  To be called if the Cron daemon does not respond or it crashed without                         */
   /*  cleaning up after itself, which has only happened to me once.                                  */
   /* =============================================================================================== */
   call GetStatus
   if pid \== '' then do
         /*  if Cron is alive, kill it:                                                               */
         if IsAlive(pid) then do
               ret = RxKillProcess(pid)
               do c = 1 to 5
                  if \IsAlive(pid) then do
                        say 'killed Cron PID' pid
                        leave c
                     end
                  call SysSleep 1
               end c
            end
         if \IsAlive(pid) then do
               call SysFileDelete pidfile
               call rxqueue 'delete', Qname
            end
         else
            /*  if Cron is still alive, it is probably better not to destroy the rest                 */
            say 'Failed to kill PID' pid
      end
   else do
         say 'cannot get PID of Cron, deleting queue'
         if IsQ then
            call rxqueue 'delete', Qname
      end
   return 0
   /* === End of Kill =============================================================================== */

   IsAlive: procedure /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        if IsAlive(pid) then ...                                                                 */
   /*  returns True if pid belongs to a running process.                                              */
   /* =============================================================================================== */
   /*  Search for PID in the list of all running processes. If found, return True                     */
   arg pid
   Alive = 0

   call RxQProcStatus 'q.'
   do i = 1 to q.0P.0
      if x2d(q.0P.i.1) == pid then do
            Alive = 1
            leave i
         end
   end i
   return Alive
   /* === End of IsAlive ============================================================================ */

   Help: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   call lineout 'stderr', G.0me 'controls the Cron daemon 'CronName
   call lineout 'stderr', ''
   call lineout 'stderr', 'You should read and possibly edit the Setup section at the top of'
   call lineout 'stderr', G.0me 'before using this.'
   call lineout 'stderr', ''
   call lineout 'stderr', 'Usage:'
   call lineout 'stderr', '   ' G.0me 'HELP | START | STOP | STATUS | PAUSE <minutes> | RESUME | DEBUG <level>'
   call lineout 'stderr', '      HELP              show this help'
   call lineout 'stderr', '      START             start the Cron.cmd daemon'
   call lineout 'stderr', '      STOP              stop the Cron.cmd daemon'
   call lineout 'stderr', '      STATUS            show status of the Cron.cmd daemon'
   call lineout 'stderr', '      PAUSE <minutes>   delay running any jobs for <minutes>'
   call lineout 'stderr', '                        delayed jobs will be run when it resumes'
   call lineout 'stderr', '      RESUME            tell Cron.cmd to resume now'
   call lineout 'stderr', '      DEBUG <level>     change the Cron.cmd debugging level'
   call lineout 'stderr', '                        0 is normal, 3 is very verbose'
   call lineout 'stderr', ''
   call lineout 'stderr', 'for testing and debugging:'
   call lineout 'stderr', '      -debug=<n>        set starting debug level, 0 to 3'
   call lineout 'stderr', '      -foreground       run Cron.cmd in a window with logging to screen'
   call lineout 'stderr', '                        read the comments in' G.0me
   call lineout 'stderr', '      KILL              kill Cron.cmd process and clean up'
   call lineout 'stderr', '                        use this only if Cron.cmd does not respond, or if it died'
   call lineout 'stderr', '                        unexpectedly without cleaning up the queue and PID file'
   call lineout 'stderr', ''
   call lineout 'stderr', 'Some values set within' G.0me 'are:'
   if symbol('CronName') == 'VAR' then
      call lineout 'stderr', '      Cron command      =' CronName
   if symbol('pidfile') == 'VAR' then
      call lineout 'stderr', '      PID file          =' pidfile
   if symbol('Qname') == 'VAR' then
      call lineout 'stderr', '      Queue name        =' Qname
   if symbol('debuglevel') == 'VAR' then
      call lineout 'stderr', '      debug level       =' debuglevel
   if symbol('crontab') == 'VAR' then do
         call lineout 'stderr', '      crontab file      =' crontab
         if \FileExist(crontab) then
            call lineout 'stderr', '                          but it does not exist'
      end
   if symbol('logdir') == 'VAR' then
      call lineout 'stderr', '      log directory     =' logdir
   if symbol('renamelog') == 'VAR' then
      call lineout 'stderr', '      rename log file   =' renamelog
   call lineout 'stderr', ''
   return
   /* === End of Help =============================================================================== */

   Mark: /*fold00*/
   0Mark:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       CALL Mark [<list of variable names>] [, <string>]                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Shows the filename and line number it was called from, the current values of the               */
   /*  variables named, and an informational string                                                   */
   /*  Arg(1) is a list of variable names                                                             */
   /*  Arg(2) can be any descriptive text                                                             */
   /*  Examples:                                                                                      */
   /*      CALL mark 'varA varB', 'At top of the "DO x = 1 to 5" loop'                                */
   /*  If no variable names are given, it will just show a message:                                   */
   /*      CALL mark '', 'At top of the "DO x = 1 to 5" loop'                                         */
   /*                                                                                                 */
   /*  This must share all the variables of whatever routine it was called from.                      */
   /*  I cannot think of a way to do it without setting at least one variable here,                   */
   /*  so reserve the variable '_' for the use of Mark and similar routines.                          */
   /* =============================================================================================== */
   _ = sigl
   call lineout 'stderr', ''
   if arg(2, 'e') ,
    & arg(2) = '' then nop
   else call lineout 'stderr', '[Mark]['value('G.0me')':'_']' arg(2)
   do _ = 1 to 99
      if word(arg(1), _) = '' then leave _
      if symbol(word(arg(1), _)) = 'VAR' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||'= "'||value(word(arg(1), _))||'"'
      if symbol(word(arg(1), _)) = 'LIT' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not defined'
      if symbol(word(arg(1), _)) = 'BAD' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not a variable'
   end _
   drop _
   return 0
   /* === End of 0Mark ============================================================================== */

   LineNo: /*fold00*/
   0LineNo:
   /* =============================================================================================== */
   /*  returns the number of the current line--    say 'now at line 'LineNo()                         */
   /* =============================================================================================== */
   _ = sigl
   return _
   /* === End of 0LineNo ============================================================================ */

   DirExist: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*      DirExist(<directory name>)                                                                 */
   /*                                                                                                 */
   /*  Examples:                                                                                      */
   /*      if DirExist('F:\os2') then ...                                                             */
   /*      if \DirExist('F:\NotThere') then ...                                                       */
   /*      if \DirExist(filespec('path', outfile)) then ...                                           */
   /*                                                                                                 */
   /*  returns 1 if a directory exists, 0 if it does not                                              */
   /*                                                                                                 */
   /*  use either \ or /                                                                              */
   /*  trailing slashes are stripped                                                                  */
   /*      'F:\OS2' or 'F:\OS2\'   both mean 'F:\OS2'                                                 */
   /*      'F:' or 'F:\'           both mean 'F:\'                                                    */
   /*                                                                                                 */
   /*  in 4os2.exe/cmd.exe,                                                                           */
   /*      'F:'                    means "the current directory on drive F:"                          */
   /*  but that would be pointless here; the current directory must exist                             */
   /*  likewise for                                                                                   */
   /*      "DirExist('.')"                                                                            */
   /*                                                                                                 */
   /*  if no directory name is provided at all, it is taken as '.'                                    */
   /*  that way a calling routine can verify the path of a filename without the                       */
   /*  extra step of checking the output of filespec() for ''.                                        */
   /*          if \DirExist(filespec('path', outfile)) then do                                        */
   /*              say 'invalid file name, no such directory'                                         */
   /*              return 1                                                                           */
   /*          end                                                                                    */
   /*                                                                                                 */
   /*  relative paths like "..\..\bin" with leading ".." are legal, but                               */
   /*  SysFileTree() can easily be confused by them                                                   */
   /*                                                                                                 */
   /*  wildcards are legal in <directory name>, but probably not useful                               */
   /* =============================================================================================== */

   /* ----------------------------------------------------------------------------------------------- */
   /*  the root directory of a drive needs special handling because SysFileTree()                     */
   /*  will not find it unless it has at least one subdirectory.                                      */
   /*                                                                                                 */
   /*  There is a logic to that--                                                                     */
   /*  to SysFileTree(), "F:\OS2" means "the directory named F:\OS2".                                 */
   /*  An ending slash, "F:\OS2\" means "all subdirectories of F:\OS2".                               */
   /*  so there is no way to refer to the root directory of a drive.                                  */
   /*  "R:\" means "all subdirectories under R:\", and                                                */
   /*  "R:" means "the current directory on R:".                                                      */
   /* ----------------------------------------------------------------------------------------------- */
   parse arg DirName .
   if DirName == '' then return 1
   if DirName == '.' then return 1
   DirName = translate(DirName, '\', '/')

   /*  if DirName = 'F:\' verify that drive F: exists. If so, it presumably has a root directory      */
   if right(DirName, 2) == ':\' then do
         DriveN = filespec('drive', DirName)
         if SysDriveInfo(DriveN) = '' then
            return 0
         else
            return 1
      end
   else do
         DirName = strip(DirName, 't', '\')
         ret = SysFileTree(DirName, 'Dirlist.', 'do')
         if Dirlist.0 > 0 then
            return 1
         else
            return 0
      end
   /* === End of DirExist =========================================================================== */

   GetOpts: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     INTERPRET GetOpts(args [,valid])                                                            */
   /*                                                                                                 */
   /*  where 'args' is the command line to be parsed, and                                             */
   /*  'valid' is an optional space-separated list of valid switches                                  */
   /*  'help' and '?' are always valid                                                                */
   /*                                                                                                 */
   /*  Example:                                                                                       */
   /*     parse arg args                                                                              */
   /*     interpret GetOpts(args, 'switchA switchB')                                                  */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Switches must start with '-'. For example:                                                     */
   /*                                                                                                 */
   /*         -Foo                (Opt.0Foo == 1)                                                     */
   /*         -Foo=1              (Opt.0Foo == 1)                                                     */
   /*         -!Foo or -Foo=0     (Opt.0Foo == 0)                                                     */
   /*         -Foo=string         (Opt.0Foo == 'string')                                              */
   /*         -Foo="string"       (Opt.0Foo == 'string')                                              */
   /*         -Foo=a,b,c          (Opt.0Foo == 'a,b,c')                                               */
   /*         -Foo="a, b, c"      (Opt.0Foo == 'a, b, c')                                             */
   /*         -Foo="a b c"        (Opt.0Foo == 'a b c')                                               */
   /*         -Foo=123            (Opt.0Foo == '123')                                                 */
   /*         -Foo= or -Foo=''    (Opt.0Foo == '')                                                    */
   /*                                                                                                 */
   /*  switches must not have spaces before or after the '='                                          */
   /*  values with spaces must be in quotes. the quotes will be stripped.                             */
   /*                                                                                                 */
   /*  a switch without a value is set to '1'                                                         */
   /*        -switchA                                                                                 */
   /*  is the same as                                                                                 */
   /*        -switchA=1                                                                               */
   /*                                                                                                 */
   /*  a switch can be negated with "!", so                                                           */
   /*        -!switchA                                                                                */
   /*  is the same as                                                                                 */
   /*        -switchA=0                                                                               */
   /*                                                                                                 */
   /*     -? is the equivalent of -help                                                               */
   /*     they will both result in:                                                                   */
   /*        Opt.0help    == 1                                                                        */
   /*                                                                                                 */
   /*  Interpreting what GetOpts returns will create several variables that the                       */
   /*  caller can use--                                                                               */
   /*     Opt.              is a stem variable that holds whatever was specified by switches          */
   /*     Opt.0optlist      a space-separated list of options that were read                          */
   /*     Opt.0errors       number of errors; that is, when GetOpts() did not understand something    */
   /*                                                                                                 */
   /*                       elements of the Opt. compound variable all start with '0', to avoid       */
   /*                       name conflicts with simple variables                                      */
   /*                                                                                                 */
   /*     ArgV.             is a stem that holds the non-switch arguments                             */
   /*                       (ArgV.0 is the name of the .cmd script, NOT the number of items           */
   /*                       in the stem)                                                              */
   /*     ArgC              is the number of non-switch arguments                                     */
   /*                                                                                                 */
   /*  Example:                                                                                       */
   /*     Foo.cmd -dir=up -msg="that way" -v "From here"                                              */
   /*     will result in these variables:                                                             */
   /*        Opt.0optlist == opt.0dir opt.0msg                                                        */
   /*        Opt.0dir     == up                                                                       */
   /*        Opt.0msg     == that way                                                                 */
   /*        Opt.0v       == 1                                                                        */
   /*        ArgV.0       == Foo.cmd                                                                  */
   /*        ArgV.1       == From here                                                                */
   /*        ArgC         == 1                                                                        */
   /*                                                                                                 */
   /*  in the case of errors, GetOpts() will set Opt.0errors to the number of errors                  */
   /*  found, and cause the caller to show a message about it on stderr. The caller                   */
   /*  should probably abort if Opt.0errors is greater than 0.                                        */
   /* =============================================================================================== */
   /* Some important variables (inside GetOpts) are:                                                  */
   /*    Str.SetOpts       == commands the the caller will execute to set options                     */
   /*                          in the Opt. stem variable                                              */
   /*    Str.SetArgs       == commands the the caller will execute to put the arguments               */
   /*                          in the ArgV. and ArgC variables                                        */
   /*    Str.Errors        == commands the caller will execute to report errors                       */
   /*                          reading the command line                                               */
   /*    ReturnStr         == all of the above, combined and returned to caller                       */
   /*    flag.             == variables for GetOpt's internal use                                     */
   /*                          flag.error == number of errors trying to read command line             */
   /*    Str.SwitchList    == list of options (switches) used on command line                         */
   /*    ArgC              == number of non-switch arguments                                          */
   /*    ArgV.             == array of non-switch arguments                                           */
   /*    valid             == list of valid switches                                                  */
   /*                                                                                                 */
   /* =============================================================================================== */
   flag. = ''
   flag.errors = 0
   str. = ''
   ReturnStr = ''
   ArgC = 0
   ArgV. = ''
   parse arg args , valid
   do i = 1 to WordsQ(args)
      thisArg = WordQ(args, i)
      if left(thisArg, 1) == '-' then
         call ReadSwitch thisArg
      else do
            /*    not a switch, add to Arg list                                                       */
            ArgC = ArgC +1
            ArgV.ArgC = thisArg
         end
   end
   /* ----------------------------------------------------------------------------------------------- */
   /*    now put it all together into a series of commands for the caller to interpret                */
   if flag.errors > 0 then do
         str.SwitchList = strip(str.SwitchList' Opt.0Errors')
      end
   str.SetOpts = str.SetOpts' Opt.0Errors = 'flag.errors ';'
   str.SetOpts = str.SetOpts 'Opt.0optlist = "'str.SwitchList'" ;'

   str.SetArgs = 'ArgV. = "" ; parse source . . ArgV.0 ;  ArgC = "'ArgC'" ;'
   do i = 1 to ArgC
      varVal = Unquoted(ArgV.i)
      str.SetArgs = str.SetArgs' ArgV.'i' = 'Quoted(varVal)' ;'
   end

   ReturnStr = str.errors str.SetOpts str.SetArgs
   return ReturnStr
   /* === End of GetOpts ============================================================================ */

   ReadSwitch: procedure expose flag. str. valid (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  this is part of GetOpts()                                                                      */
   /* =============================================================================================== */
   parse arg thisArg
   Negate = 0
   switch = strip(thisArg, 'L', '-')
   if left(switch, 1) == '!' then do
         switch = strip(switch, 'L', '!')
         Negate = 1
      end
   parse var switch optName '=' optVal
   select
      /* -? and -help have fixed meanings and are allowed even if not included in 'valid'             */
      when translate(optName) == '?' then
         optName  = 'Opt.0Help'
      when translate(optName) == 'HELP' then
         optName  = 'Opt.0Help'
      when optName == '' then do
            str.errors = ,
              str.errors||'call lineout "stderr", "[GetOpts] `-` without a switch name" ;'
            flag.errors = flag.errors +1
            return
         end
      when valid \= '' ,
         & wordpos(translate(optName), translate(valid)) == 0 then do
            str.errors = str.errors|| ,
                         'call lineout "stderr", "[GetOpts] I do not understand switch `'thisArg'`" ;'
            flag.errors = flag.errors +1
            return
         end
      otherwise
         optName = 'Opt.0'optName
   end   /*  select  */

   select
      when Negate then
         if optVal == '' then
            optVal = 0
         else do
               str.errors = str.errors|| ,
                            'call lineout "stderr", "[GetOpts] I do not understand switch `'thisArg'`" ;'
               flag.errors = flag.errors +1
               return
            end
      when optVal == '' ,
         & right(switch, 1) \= '=' then
         /*  "-switch=" means Opt.0switch == '', not 1                                                */
         optVal = '1'
      otherwise
         /*    try to be tolerant of unusual arguments; only strip quotes if they are balanced        */
         /*    and only one level of nested quotes                                                    */
         optVal = Unquoted(optVal)
   end   /*  select  */

   if wordpos(optName, str.SetOpts) > 0 then do
         str.errors = ,
           str.errors||'call lineout "stderr", "[GetOpts] duplicate switch `'thisArg'`" ;'
         flag.errors = flag.errors +1
      end
   else do
         str.SetOpts = str.SetOpts' 'optName' = 'Quoted(optVal)' ;'
         str.SwitchList = strip(str.SwitchList' 'optName)
      end
   return
   /* === End of ReadSwitch ========================================================================= */

   FileExist: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        if FileExist(<filename>) then                                                            */
   /*                                                                                                 */
   /*  returns 1 (True) if <filename> exists and 0 (False) if it does not.                            */
   /*                                                                                                 */
   /*  <filename> must be a real file. This returns 0 for devices or for streams like STDOUT.         */
   /*  Granted, STDOUT does exist, but it is more useful this way. I mostly use this when I want      */
   /*  to avoid clobbering an existing file,                                                          */
   /*                                                                                                 */
   /*  The main reason for this is that                                                               */
   /*        if FileExist(filename) then                                                              */
   /*  is easier to remember than                                                                     */
   /*         if stream(filename, 'c', 'query exists') then                                           */
   /* =============================================================================================== */
   /*  Most of the complications are because OS/2 Rexx and Regina return different values             */
   /*  from stream() for stdout and for devices. Regina usually returns 0 for stdout, but             */
   /*  not always. I don't know why.                                                                  */
   parse arg FileN
   if FileN = '' then return 0
   if wordpos(translate(FileN), 'STDOUT STDERR') > 0 then return 0
   ret = stream(FileN, 'c', 'query exists')
   if ret == '' then return 0
   if left(translate(ret), 5) == '\DEV\' then return 0
   if left(translate(ret), 5) == '/DEV/' then return 0
   return 1
   /* === End of FileExist ========================================================================== */

   DayOfWeek: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        day = DayOfWeek(datestring)                                                              */
   /*        day = DayOfWeek('2019-01-31')                                                            */
   /*        day = DayOfWeek('2019/1/31')                                                             */
   /*  returns the day of week for a given date, as a string. like 'Monday'                           */
   /*  the date =must= be in YYYY MM DD order                                                         */
   /*  the year =must= have 4 digits; it does not make any assumptions about centuries.               */
   /*  other than that it is flexible--                                                               */
   /*        the month and day can be 1 or 2 digits,                                                  */
   /*        the separator character can be ' ', '-', '/', or practically anything except a number.   */
   /*        anything after the date is ignored, so you do not have to remove a time, like            */
   /*                  DayOfWeek('2019-9-1 03:34:30')                                                 */
   /*        but there must be a space after the date                                                 */
   /*                                                                                                 */
   /* =============================================================================================== */
   /*  Slightly changed from the Zeller's congruence procedure in                                     */
   /*      Teach Yourself REXX in 21 Days by William F. Schindler and Esther Schindler                */
   /*                                                                                                 */
   /*    added by akm: allow any separator between numbers                                            */
   /*                  return day name instead of number                                              */
   /*                  some error checking                                                            */
   /* =============================================================================================== */
   parse arg datestring
   sep = substr(datestring, 5, 1)
   parse var datestring year (sep) month (sep) day .
   if length(year) \= 4 then return ''
   if datatype(year, 'W') == 0 then return ''
   if datatype(month, 'W') == 0 then return ''
   if datatype(day, 'W') == 0 then return ''
   /* =============================================================================================== */
   if month > 2 then do
         adjMonth = month -2
         adjYear = year
      end
   else do
         adjMonth = month +10
         adjYear = year -1
      end

   century = adjYear % 100
   yearInCentury = adjYear -100 * century
   yrAdj = yearInCentury + (yearInCentury % 4) + (century % 4)
   yrAdj = yrAdj - century - century +77
   dayOfWeek = ((13 * adjMonth -1) % 5) + day + yrAdj
   dayOfWeek = dayOfWeek // 7

   DowN = word('Sunday Monday Tuesday Wednesday Thursday Friday Saturday Sunday', dayOfWeek +1)
   return DowN
   /* === End of DayOfWeek ========================================================================== */

   WordsQ: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordsQ(string, <separator characters>, <quote characters>)                            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a drop-in replacement for the Rexx Word() function.                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Words() function--                                                                    */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg str , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(str) ,
          & pos(substr(str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(str) then numwords = numwords +1

   do while (i <= length(str))
      if pos(substr(str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(str) ,
                   & substr(str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(str, i, 1), chrsW) \= 0 ,
       & i <= length(str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            do while (pos(substr(str, i, 1), chrsW) \= 0) ,
                   & (i <= length(str))
               i = i +1
            end
            if i <= length(str) then numwords = numwords +1
         end
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   return numwords
   /* === End of WordsQ ============================================================================= */

   WordQ: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordQ(<string>, <word number>, <separator characters>, <quote characters>)            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a replacement for the Rexx Word() function.                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Word() function--                                                                     */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg str , thisnum , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0
   Thisword  = ''
   w1        = 0
   w2        = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(str) ,
          & pos(substr(str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(str) then numwords = numwords +1
   if numwords == thisnum then w1 = i

   do while (i <= length(str))
      if pos(substr(str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(str) ,
                   & substr(str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(str, i, 1), chrsW) \= 0 ,
       & i <= length(str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            /* unless the end of the word we want, then quit                                          */
            if w1 > 0 then do
                  w2 = i -1
                  leave
               end
            do while (pos(substr(str, i, 1), chrsW) \= 0) ,
                   & (i <= length(str))
               i = i +1
            end
            if i <= length(str) then numwords = numwords +1
            /* if this is the start of the word we want, note position in string                      */
            if numwords == thisnum then w1 = i
         end   /*  wordsep  */
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   if i > length(str) then i = length(str)
   if w1 > 0 then do
         /* we know w1, if we ran out of string, w2 is the end of string                              */
         if w2 == 0 then w2 = i
         /* found start and end of word, copy it                                                      */
         wlen = w2 - w1 +1
         Thisword = substr(str, w1, wlen)
      end
   return Thisword
   /* === End of WordQ ============================================================================== */

   Quoted: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Quoted(string)                                                                  */
   /*                                                                                                 */
   /*    if possible, quotes a string with the kind of quotes it does not contain                     */
   /*        {He said "no"}   --> {'He said "no"'}                                                    */
   /*        {He said 'no'}   --> {"He said 'no'"}                                                    */
   /*    otherwise uses double quotes and lets the caller deal with the errors                        */
   /* =============================================================================================== */
   parse arg string
   select
      when pos('"', string) == 0 then
         newstring = '"'string'"'
      when pos("'", string) == 0 then
         newstring = "'"string"'"
      otherwise
         newstring = '"'string'"'
   end
   return newstring
   /* === End of Quoted ============================================================================= */

   ShellName: procedure /*fold00*/
   /* =============================================================================================== */
   /*  returns "4OS2" or "CMD" or "UNKNOWN" depending on the shell/command processor being used       */
   /* =============================================================================================== */
   /*  4OS2.EXE: line 2 is "4OS2   3.09    OS/2 Version is 4.50"                                      */
   /*  CMD.EXE   line 2 is "The Operating System/2 Version is 4.50"                                   */
   Qstart = queued()
   address cmd '@ver | rxqueue.exe /lifo'
   parse pull shellname
   /*  there should be one more line. Put the queue back where it started                             */
   do while queued() > Qstart
      parse pull dummy
   end
   if word(shellname, 1) == '4OS2' then return '4OS2'
   if abbrev(shellname, 'The Operating System/2') then return 'CMD'
   return 'UNKNOWN'
   /* === End of ShellName ========================================================================== */

   Unquoted: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Unquoted(string)                                                                */
   /*                                                                                                 */
   /*    strips quotes from a string if and only if it is enclosed in a balanced pair                 */
   /*    of single or double quotes                                                                   */
   /* =============================================================================================== */
   parse arg string
   if length(string) < 2 then return string
   select
      when left(string, 1) == '"' ,
         & right(string, 1) == '"' then
         string = substr(string, 2, length(string) -2)
      when left(string, 1) == "'" ,
         & right(string, 1) == "'" then
         string = substr(string, 2, length(string) -2)
      otherwise
         nop
   end
   return string
   /* === End of Unquoted =========================================================================== */

   NoValue: /*fold00*/
   /* =============================================================================================== */
   signal off NoValue
   NVsigl = sigl
   call lineout 'stderr', G.0me '[NOVALUE Handler] at line 'NVsigl' for variable "'condition('d')'"'
   call lineout 'stderr', '   ==> {'sourceline(NVsigl)'}'
   exit 9
   /* === End of NoValue ============================================================================ */


